from .EntropyBasedMethod import EntropyBasedMethod
from .moving_window import MovingWindow
from .Static import Static
from .SubsamplingMethod import SubsamplingMethod
from .WSDynamic import WSDynamic
from .WSFFTStatic import WSFFTStatic
from .WSStatic import WSStatic