from fpcmci.version import VERSION

"""
FPCMCI.

Filtered - PCMCI causal discovery algorithm.
Extension of the PCMCI causal discovery algorithm augmented with a feature selection method.
"""

__version__ = VERSION
__author__ = 'Luca Castri'